/*
 * StrawLib v0.2
 *
 * Copyright (c) 2005 Pr Spjuth
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and
 * associated documentation files (the "Software"), to deal in the Software without restriction, 
 * including without limitation the rights to use, copy, modify, merge, publish, distribute, 
 * sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or 
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT
 * NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/* StrawLib Headers */
#include "straw.h"


/**************************************************************************************************
 * Constants
 **************************************************************************************************/

/**
 * These standard constants have been defined for readability reasons.
 */
#define TRUE   (-1)
#define FALSE  0
#define NULL   0


/**************************************************************************************************
 * Local Functions
 **************************************************************************************************/

/**
 * Appends a new task to the task queue. Does however make an exception for termination tasks, which
 * are automatically placed at the beginning of the queue.
 *
 * @param  Session_p  Pointer to session to which the task queue belongs.
 * @param  Action     The action that should be taken when processing the task.
 * @param  Data_p     Pointer to data needed to perform the task.
 */
static void
AppendTask(StrawSession_t*   Session_p, 
           StrawTaskAction_t Action, 
           void*             Data_p)
{
  StrawTask_t *NewTask_p = NULL;

  /* Set task */
  NewTask_p         = (StrawTask_t*)Session_p->AllocMem(sizeof(StrawTask_t));
  NewTask_p->Action = Action;
  NewTask_p->Data_p = Data_p;

#ifdef STRAW_CFG_DEBUG
  /* Update memory overhead */
  Session_p->Overhead += sizeof(StrawTask_t);
  if (Session_p->Overhead > Session_p->MaxOverhead) {
    Session_p->MaxOverhead = Session_p->Overhead;
  }
#endif

  /* Terminations must be placed before previous starts, to stop new children before they
   * start execute.
   */
  if (Action == STRAW_TASKTERMINATESTRAW) {
    /* Handle special case with empty list */
    if (Session_p->TaskQueueHead_p == NULL) {
      Session_p->TaskQueueHead_p = NewTask_p;
      NewTask_p->NextTask_p = NULL;
    }
    else {
      Session_p->TaskQueueTail_p->PrevTask_p = NewTask_p;
      NewTask_p->NextTask_p = Session_p->TaskQueueTail_p;
    }

    /* Prepend at beginning of queue */
    NewTask_p->PrevTask_p = NULL;
    Session_p->TaskQueueTail_p = NewTask_p;
  }
  else {
    /* Handle special case with empty list */
    if (Session_p->TaskQueueTail_p == NULL) {
      Session_p->TaskQueueTail_p = NewTask_p;
      NewTask_p->PrevTask_p = NULL;
    }
    else {
      Session_p->TaskQueueHead_p->NextTask_p = NewTask_p;
      NewTask_p->PrevTask_p = Session_p->TaskQueueHead_p;
    }

    /* Append at end of queue */
    NewTask_p->NextTask_p = NULL;  
    Session_p->TaskQueueHead_p = NewTask_p;
  }
}


/**
 * Removes the specified task from the task queue. This function has to be used to ensure the
 * integrity of the double linked list making up the queue.
 *
 * @param  Session_p  Pointer to session to which the task queue belongs.
 * @param  Task_p     Pointer to task to be removed from the queue.
 */
static void
RemoveTask(StrawSession_t *const Session_p, 
           StrawTask_t*          Task_p)
{  
  if (Task_p == Session_p->TaskQueueHead_p) {
    /* Special case when the task to remove is the tail */
    Session_p->TaskQueueHead_p = Task_p->PrevTask_p;
  }
  else {
    /* If it was not the tail then we have to fix the link at the next task */
    Task_p->NextTask_p->PrevTask_p = Task_p->PrevTask_p;
  }
  if (Task_p == Session_p->TaskQueueTail_p) {
    /* Special case when the task to remove is the head */
    Session_p->TaskQueueTail_p = Task_p->NextTask_p;
  }
  else {
    /* If it was not the head then we have to fix the link at the previous task */
    Task_p->PrevTask_p->NextTask_p = Task_p->NextTask_p;
  }
#ifdef STRAW_CFG_DEBUG
  Session_p->Overhead -= sizeof(StrawTask_t);
#endif
  Session_p->FreeMem(Task_p);
}


static unsigned int
CalcualteSignalHash(unsigned int    InternalSignal,
                    unsigned int    ExternalSignal)
{
  if (ExternalSignal > 0) {
    return ExternalSignal & 0xf;
  }
  else {
    return InternalSignal & 0xf;
  }
}


/**
 * Allocates and initializes a new fresh straw.
 *
 * @param  Session_p   Pointer to the session to which the straw will belong;
 * @param  EntryPoint  Chunk function to be executed when this straw is run.
 *
 * @returns  Pointer to the new straw.
 */
static Straw_t*
CreateStraw(StrawSession_t* Session_p, 
            StrawChunk_t    EntryPoint,
            unsigned int    PrivateDataSize)
{
  Straw_t *NewStraw_p = NULL;

  /* Initialize instance */
  NewStraw_p                     = (Straw_t*)Session_p->AllocMem(sizeof(Straw_t) + PrivateDataSize);
  NewStraw_p->SlotB.CurrentChunk = EntryPoint;
  NewStraw_p->SlotA.CleanupChunk = NULL;
  NewStraw_p->ParentStraw_p      = NULL;
  NewStraw_p->NextStraw_p        = NULL;
  NewStraw_p->Flags              = 0;
#ifdef STRAW_CFG_DEBUG
  NewStraw_p->Label_p            = NULL;
  NewStraw_p->CleanupLabel_p     = NULL;

  /* Update memory overhead */
  Session_p->Overhead += sizeof(Straw_t);
  if (Session_p->Overhead > Session_p->MaxOverhead) {
    Session_p->MaxOverhead = Session_p->Overhead;
  }
#endif

  return NewStraw_p;
}


/**
 * 
 */
static void
RemoveFromHash(StrawSession_t* Session_p, 
               Straw_t *const  Straw_p)
{
  /* Only remove if the straw is really in the hash */
  if (Straw_p->Flags & STRAW_WAITFORSIGNAL) {
    unsigned int SignalHash;

    /* Get signal hash */
    SignalHash = CalcualteSignalHash(Straw_p->SlotC.SignalMapping_p[0].InternalSignal, 
                                     Straw_p->SlotC.SignalMapping_p[0].ExternalSignal);

    /* Handle special case where entry is first in hash list */
    if (Session_p->MappingsHash_p[SignalHash] == Straw_p) {
      Session_p->MappingsHash_p[SignalHash] = Straw_p->SlotB.NextHashEntry_p;
    }
    else {
      Straw_t *HashEntry_p = Session_p->MappingsHash_p[SignalHash];

      /* Scan hash for current straw */
      while (HashEntry_p->SlotB.NextHashEntry_p != Straw_p) {
        HashEntry_p = HashEntry_p->SlotB.NextHashEntry_p;
      }

      /* Relink hash */
      HashEntry_p->SlotB.NextHashEntry_p = Straw_p->SlotB.NextHashEntry_p;
    }

    Straw_p->Flags &= ~STRAW_WAITFORSIGNAL;
  }
}


/**
 * 
 */
/* NOTE: The provided straw is not always the current straw, why both parameters are needed. */
static void
TerminateStraw(StrawSession_t* Session_p, 
               Straw_t *const  Straw_p)
{
  /* This straw should no longer be notified of when a terminated straw has finished */
  if (Straw_p->Flags & STRAW_WAITFORTERMINATION) {
    Straw_t *Terminator_p = Straw_p;

    /* Find the terminator which knows what straw is waited for */
    while (!(Terminator_p->Flags & STRAW_LASTTERMINATOR)) {
      Terminator_p = Terminator_p->SlotC.NextTerminator_p;
    }

    /* Remove this straw from terminator list */
    if (Straw_p == Terminator_p->SlotC.TerminatedStraw_p->SlotA.Terminator_p) {
      Terminator_p->SlotC.TerminatedStraw_p->Flags &= ~STRAW_HASTERMINATOR;      
    }
    else {
      /* Straw searching for the terminator before this straw */
      Terminator_p = Terminator_p->SlotC.TerminatedStraw_p->SlotA.Terminator_p;
      while (Terminator_p->SlotC.NextTerminator_p != Straw_p) {
        Terminator_p = Terminator_p->SlotC.NextTerminator_p;
      }
      Terminator_p->SlotC.NextTerminator_p = Straw_p->SlotC.NextTerminator_p;
    }
  }

  /* Make sure this straw is marked as terminating, so child/sub-straws can act on it */
  StrawLogTerminationAttempt(Straw_p);
  Straw_p->Flags |= STRAW_TERMINATING;

  /* Make sure the straw is removed from the hash table */
  RemoveFromHash(Session_p,Straw_p);

  if (Straw_p->Flags & STRAW_RUNNINGSUB) {
    StrawLogTerminateChildStraw(Straw_p, Straw_p->SlotC.SubStraw_p, TRUE);
    AppendTask(Session_p, STRAW_TASKTERMINATESTRAW, Straw_p->SlotC.SubStraw_p);

    /* Wait for sub-straw to finish */
    return;
  }
  else if (Straw_p->Flags & STRAW_PARENT) {
    Straw_t *ChildStraw_p = Session_p->StrawList_p;
  
    /* Terminate all children for this straw */
    /* NOTE: The children cannot be terminated directly here because that could potentially mess up
     *       the view of the straw list we got here, and it could also lead to large call stacks.
     *       Therefore the terminations are postponed and queued.
     */
    Straw_p->SlotC.ChildCount = 0;
    while (ChildStraw_p != NULL) {
      if (ChildStraw_p->ParentStraw_p == Straw_p) {
        StrawLogTerminateChildStraw(Straw_p, ChildStraw_p, FALSE);
        AppendTask(Session_p, STRAW_TASKTERMINATESTRAW, ChildStraw_p);

        /* Remember this child */
        Straw_p->SlotC.ChildCount++;
      }
      ChildStraw_p = ChildStraw_p->NextStraw_p;
    }

    /* Wait for children to finish */
    /* NOTE: Because a parent never knows how many children still exists all of them may have
     *       already terminated.
     */
    if (Straw_p->SlotC.ChildCount > 0) {
      return;
    }
  }

  if (Straw_p->SlotB.CurrentChunk != NULL) {    
    StrawLogRunCleanup(Straw_p, Straw_p->CleanupLabel_p);

    /* Run cleanup state before removing straw */
    AppendTask(Session_p, STRAW_TASKRUNSTRAW, Straw_p);
  }
  else {
    StrawTask_t *Task_p = Session_p->TaskQueueTail_p;

    /* Remove straw from global list */
    if (Session_p->StrawList_p == Straw_p) {
      Session_p->StrawList_p = Straw_p->NextStraw_p;
    }
    else {
      Straw_t *PrevStraw_p = NULL;
      
      /* Search for previous straw to be able to correctly remove this one */
      /* NOTE: Speed has been sacrificed here since we otherwise would have to have a
      *        double linked list to hold this.
      */
      PrevStraw_p = Session_p->StrawList_p;
      while (PrevStraw_p->NextStraw_p != Straw_p) {
        PrevStraw_p = PrevStraw_p->NextStraw_p;
      }
      PrevStraw_p->NextStraw_p = Straw_p->NextStraw_p;
    }

    /* Attempt to terminate the parent again if was waiting for children to terminate */
    if (Straw_p->ParentStraw_p != NULL) {     
      if (Straw_p->ParentStraw_p->Flags & STRAW_TERMINATING) {
        if (Straw_p->Flags & STRAW_SUBSTRAW) {
          Straw_p->ParentStraw_p->Flags &= ~STRAW_RUNNINGSUB;
          AppendTask(Session_p, STRAW_TASKTERMINATESTRAW, Straw_p->ParentStraw_p);
        }
        else {
          /* Remove this straw from parents waiting count */
          Straw_p->ParentStraw_p->SlotC.ChildCount--;

          /* Try terminating parent again when all children have finished */
          if (Straw_p->ParentStraw_p->SlotC.ChildCount == 0) {
            Straw_p->ParentStraw_p->Flags &= ~STRAW_PARENT;
            AppendTask(Session_p, STRAW_TASKTERMINATESTRAW, Straw_p->ParentStraw_p);
          }
        }
      }
      else if (Straw_p->Flags & STRAW_SUBSTRAW) {
        /* Resume parent if this was a sub-straw */
        Straw_p->ParentStraw_p->Flags &= ~STRAW_RUNNINGSUB;
        AppendTask(Session_p, STRAW_TASKRUNSTRAW, Straw_p->ParentStraw_p);
      }
    }

    /* Resume all terminators */
    if (Straw_p->Flags & STRAW_HASTERMINATOR) {
      Straw_t *Terminator_p = Straw_p->SlotA.Terminator_p;

      while (Terminator_p != NULL) {
        Terminator_p->Flags &= ~STRAW_WAITFORTERMINATION;
        AppendTask(Session_p, STRAW_TASKRUNSTRAW, Terminator_p);

        /* Stop when last terminator reached */
        if (Terminator_p->Flags & STRAW_LASTTERMINATOR) {
          Terminator_p->Flags &= ~STRAW_LASTTERMINATOR;
          Terminator_p = NULL;
        }
        else {
          Terminator_p = Terminator_p->SlotC.NextTerminator_p;
        }
      }
    }

    /* Remove all dependent tasks */
    while (Task_p != NULL) {
      StrawTask_t *ThisTask_p = Task_p;

      Task_p = Task_p->NextTask_p;
      if (ThisTask_p->Data_p == Straw_p) {
        RemoveTask(Session_p, ThisTask_p);
      }      
    }

#ifdef STRAW_CFG_DEBUG
    //StrawLog(Straw_p, /*STRAW_CFG_LOGINDENT*/ "Straw deleted");
    StrawLogDeleted(Straw_p);
    Session_p->Overhead -= sizeof(Straw_t);
#endif

    /* Free straw instance including any private data */
    Session_p->FreeMem(Straw_p);
  }
}


static int
ExecuteStraw(StrawSession_t* Session_p, 
             Straw_t*        Straw_p)
{
  StrawChunkAction_t NextAction;

  /* Run straw as long as it is synchronous */
  Session_p->CurrentStraw_p = Straw_p;
  do {
    /* Execute the current chunk */
    NextAction = Straw_p->SlotB.CurrentChunk(Session_p);

    /* Signal is only available for first chunk */
    Session_p->SignalType   = 0;
    Session_p->SignalData_p = NULL;

  } while (NextAction == CHUNK_NEXTCHUNK && NextAction != CHUNK_YIELD);
  
  /* Act on reason for exiting sequence */
  if (NextAction == CHUNK_NEXTCHUNK) {
    AppendTask(Session_p, STRAW_TASKRUNSTRAW, Session_p->CurrentStraw_p);
  }
  else if (NextAction == CHUNK_TERMINATE) {
    StrawLogExit(Session_p->CurrentStraw_p);

    /* Must run cleanup chunk if not already runing termination sequence */
    if (!(Straw_p->Flags & STRAW_TERMINATING)) {
      Straw_p->SlotB.CurrentChunk = Straw_p->SlotA.CleanupChunk;
    }
    TerminateStraw(Session_p, Straw_p);
  }
  else if (NextAction == CHUNK_YIELD) {
    AppendTask(Session_p, STRAW_TASKRUNSTRAW, Session_p->CurrentStraw_p);
    return TRUE;
  }

  return FALSE;
}


static int
ExecuteMappedStraw(StrawSession_t* Session_p, 
                   unsigned int    InternalSignal,
                   unsigned int    ExternalSignal,
                   void*           SignalData_p)
{
  int          FirstTry        = TRUE;
  Straw_t*     LastMatchPrev_p = NULL;
  Straw_t*     LastMatch_p     = NULL;
  unsigned int LastMatchSlot;
  unsigned int SignalHash;
  unsigned int Slot;

  /* Start at hashed slot */
  SignalHash = CalcualteSignalHash(InternalSignal, ExternalSignal);
  Slot       = SignalHash;

  /* Map signal agains entry in hash */
  while (Slot < 16) {
    Straw_t *PrevEntry_p = NULL;
    Straw_t *Entry_p     = Session_p->MappingsHash_p[Slot];

    while (Entry_p != NULL) {
      const StrawSignalMapping_t *Mapping_p   = NULL;
      int                         MappingSlot = 0;

      /* Skip first mapping of the hash since it was already scanned on the first try */
      if (!FirstTry && Slot == SignalHash) {
        MappingSlot = 1;
      }

      /* Check all the entries of the associated mapping */
      Mapping_p = &Entry_p->SlotC.SignalMapping_p[MappingSlot];
      while (Mapping_p->Chunk != NULL) {
        if (Mapping_p->InternalSignal == InternalSignal &&
            Mapping_p->ExternalSignal == ExternalSignal &&
            (Mapping_p->Matcher == NULL || 
             Mapping_p->Matcher((unsigned char*)Entry_p + sizeof(Straw_t), SignalData_p))) {
          if (Mapping_p->Matcher == StrawMatchDropped) {
            /* Save match until later when the signal may have been dropped */
            LastMatchPrev_p = PrevEntry_p;
            LastMatch_p     = Entry_p;
            LastMatchSlot   = Slot;
            LastMatch_p->SlotB.CurrentChunk = Mapping_p->Chunk;
          }
          else {
            /* Set chunk corresponding to signal */
            Entry_p->SlotB.CurrentChunk = Mapping_p->Chunk;

            /* Remove all mappings for this straw */
            if (PrevEntry_p == NULL) {
              Session_p->MappingsHash_p[Slot] = Entry_p->SlotB.NextHashEntry_p;
            }
            else {
              PrevEntry_p->SlotB.NextHashEntry_p = Entry_p->SlotB.NextHashEntry_p;
            }

            /* Execute straw */
            Entry_p->Flags &= ~STRAW_WAITFORSIGNAL;
            ExecuteStraw(Session_p, Entry_p);

            return TRUE;
          }
        }

        /* Abort matching early if we are currently only interested in first mapping */
        if (FirstTry) {
          break;
        }

        /* Now try next mapping */
        MappingSlot++;
        Mapping_p = &Entry_p->SlotC.SignalMapping_p[MappingSlot];
      }

      /* Signal matching continues in the next hash entry */
      PrevEntry_p = Entry_p;
      Entry_p = Entry_p->SlotB.NextHashEntry_p;            
    }

    /* Change matching environment after first try */
    if (FirstTry) {
      FirstTry = FALSE;
      Slot = 0;
    }
    else {
      Slot++;
    }
  }

  /* Use last drop-mapping if signal was dropped */
  if (LastMatch_p != NULL) {
    /* Remove all mappings for this straw */
    if (LastMatchPrev_p == NULL) {
      Session_p->MappingsHash_p[LastMatchSlot] = LastMatch_p->SlotB.NextHashEntry_p;
    }
    else {
      LastMatchPrev_p->SlotB.NextHashEntry_p = LastMatch_p->SlotB.NextHashEntry_p;
    }

    /* Execute straw */
    LastMatch_p->Flags &= ~STRAW_WAITFORSIGNAL;
    ExecuteStraw(Session_p, LastMatch_p);

    return TRUE;
  }

  return FALSE;
}


/**************************************************************************************************
 * Global Functions
 **************************************************************************************************/

int
StrawMatchAll(void* StrawData_p, 
              void* SignalData_p)
{
  return TRUE;
}


int
StrawMatchDropped(void* StrawData_p, 
                  void* SignalData_p)
{
  /* NOTE: This result will however only affect what straw should be remembered if no one else
   *       caught the signal.
   */
  return TRUE;
}


void
StrawCreateSession(StrawSession_t* Session_p,
                   StrawAlloc_t    AllocMem,
                   StrawFree_t     FreeMem)
{
  int i;

  /* Store memory functions */
  Session_p->AllocMem = AllocMem;
  Session_p->FreeMem = FreeMem;

  /* Initialize the task queue */
  Session_p->TaskQueueTail_p = NULL;
  Session_p->TaskQueueHead_p = NULL;

  /* Start with no mappings */
  for (i = 0; i < 16; i++) {
    Session_p->MappingsHash_p[i] = NULL;
  }

#ifdef STRAW_CFG_DEBUG
  /* Initialize memory overhead */
  Session_p->Overhead    = sizeof(StrawSession_t);
  Session_p->MaxOverhead = Session_p->Overhead;
#endif

  /* No triggering signal to start with */
  Session_p->SignalType   = 0;
  Session_p->SignalData_p = NULL;

  /* Initialize straw list to properly be able to start mother straw*/
  Session_p->StrawList_p    = NULL;
  Session_p->CurrentStraw_p = NULL;
}


StrawSessionStatus_t
StrawSessionProcess(StrawSession_t* Session_p,
                    unsigned int    InternalSignal,
                    unsigned int    ExternalSignal,
                    void*           SignalData_p,
                    int*            SignalConsumed_p)
{
  Straw_t           *Straw_p  = NULL;
  int                Consumed = FALSE;
  int                Yielding = FALSE;
  StrawTaskAction_t  Action;

  /* Processing signal comes last, to give children a chance to start */
  if (InternalSignal != 0 || ExternalSignal != 0) {
    AppendTask(Session_p, STRAW_TASKMAPSIGNAL, NULL);
  }

  /* Carry out all tasks, up to a yield */
  while (!Yielding && Session_p->TaskQueueTail_p != NULL) {

    /* Pick task with highest priority */
    Action = Session_p->TaskQueueTail_p->Action;
    Straw_p = (Straw_t*)Session_p->TaskQueueTail_p->Data_p;  
    RemoveTask(Session_p, Session_p->TaskQueueTail_p);
    
    /* Process task */
    switch (Action) {
      case STRAW_TASKMAPSIGNAL:
        Consumed = ExecuteMappedStraw(Session_p, InternalSignal, ExternalSignal, SignalData_p);
        break;
      case STRAW_TASKRUNSTRAW:
        Yielding = ExecuteStraw(Session_p, Straw_p);
        break;
      case STRAW_TASKTERMINATESTRAW:
        /* Must run cleanup chunk if not already runing termination sequence */
        if (!(Straw_p->Flags & STRAW_TERMINATING)) {
          Straw_p->SlotB.CurrentChunk = Straw_p->SlotA.CleanupChunk;
        }
        TerminateStraw(Session_p, Straw_p);
        break;
      default:
        /* Discard unknown task */
        break;
    }
  }

  /* Report if signal was consumed */
  if (Consumed && SignalConsumed_p != NULL) {
    *SignalConsumed_p = Consumed;
  }

  /* Indicate session status */
  if (Yielding) {
    return STRAW_SESSIONACTIVE;
  }
  else if (Session_p->StrawList_p == NULL) {
    return STRAW_SESSIONFINISHED;
  }

  return STRAW_SESSIONWAITING;
}


Straw_t*
StrawStartChild(StrawSession_t* Session_p,
                StrawChunk_t    EntryPoint, 
                StrawChunk_t    ReturnChunk,
#ifdef STRAW_CFG_DEBUG
                const char*     ReturnLabel_p,
#endif
                const void*     PrivateData_p,
                unsigned int    PrivateDataSize)
{
  Straw_t             *CurrentStraw_p = Session_p->CurrentStraw_p;
  Straw_t             *NewStraw_p     = NULL;
  unsigned char       *Dest_p         = NULL;
  const unsigned char *Source_p       = NULL;
  unsigned int         i;

  /* Initialize new straw */
  NewStraw_p = CreateStraw(Session_p, EntryPoint, PrivateDataSize);

  /* Initialize private data by copying parameters*/
  /* NOTE: This is basically just a memcpy, but we do not use that function to minimize library
   *       dependencies.
   */
  if (PrivateData_p != NULL) {
    Dest_p = (unsigned char*)NewStraw_p + sizeof(Straw_t);
    Source_p = PrivateData_p;
    for (i = 0; i < PrivateDataSize; i++) {
      Dest_p[i] = Source_p[i];
    }
  }

  /* Establish relationships */
  NewStraw_p->ParentStraw_p = CurrentStraw_p;
  if (ReturnChunk != NULL) {
    NewStraw_p->Flags |= STRAW_SUBSTRAW;
    CurrentStraw_p->Flags |= STRAW_RUNNINGSUB;
    CurrentStraw_p->SlotC.SubStraw_p = NewStraw_p;
    CurrentStraw_p->SlotB.CurrentChunk = ReturnChunk;
  }
  else if (CurrentStraw_p != NULL) {
    CurrentStraw_p->Flags |= STRAW_PARENT;
  }

  /* Add straw to global list */
  /* NOTE: Important assumption here is that because of the insertion order, and because the parent
   *       suspends when running a sub-straw, we always know that the first encountered child for
   *       a parent is the sub-straw.
   */
  NewStraw_p->NextStraw_p = Session_p->StrawList_p;
  Session_p->StrawList_p = NewStraw_p;

  /* Schedule the new straw to be run after current chunk */
  AppendTask(Session_p, STRAW_TASKRUNSTRAW, NewStraw_p);

#ifdef STRAW_CFG_DEBUG
  if (ReturnChunk == NULL) {
    if (CurrentStraw_p != NULL) {
      StrawLogRunChild(Session_p->CurrentStraw_p, NewStraw_p);
    }
  }
  else {
    StrawLogRunSub(Session_p->CurrentStraw_p, NewStraw_p, ReturnLabel_p);
  }
#endif

  return NewStraw_p;
}


int
StrawTerminate(StrawSession_t* Session_p, 
               Straw_t*        Straw_p, 
               StrawChunk_t    ReturnChunk_p)
{
  Straw_t *RealStraw_p = Session_p->StrawList_p;

  /* Set return straw */
  Session_p->CurrentStraw_p->SlotB.CurrentChunk = ReturnChunk_p;

  /* Make sure the straw exists */
  while (RealStraw_p != NULL) {
    if (RealStraw_p == Straw_p) {
      break;
    }
    RealStraw_p = RealStraw_p ->NextStraw_p;
  }
  if (RealStraw_p == NULL) {
    return FALSE;
  }

  /* Mark the terminator */
  Session_p->CurrentStraw_p->Flags |= STRAW_WAITFORTERMINATION;

  /* Remove the hash association already here, since we need that field in the straw data for
   * storing the list of terminators.
   */
  RemoveFromHash(Session_p, Straw_p);

  /* Add to list of terminators */
  Session_p->CurrentStraw_p->SlotC.NextTerminator_p = NULL;
  if (Straw_p->Flags & STRAW_HASTERMINATOR) {
    Session_p->CurrentStraw_p->SlotC.NextTerminator_p = Straw_p->SlotA.Terminator_p;
    Straw_p->SlotA.Terminator_p = Session_p->CurrentStraw_p;
  }
  else {
    /* Set the cleanup chunk so we can use that space to store the first terminator */
    Straw_p->SlotB.CurrentChunk = Straw_p->SlotA.CleanupChunk;
    Straw_p->SlotA.Terminator_p = Session_p->CurrentStraw_p;
    Straw_p->Flags |= STRAW_HASTERMINATOR;

    /* The last terminator holds pointer to the straw waited for */
    Session_p->CurrentStraw_p->Flags |= STRAW_LASTTERMINATOR;
    Session_p->CurrentStraw_p->SlotC.TerminatedStraw_p = Straw_p;

    /* Perform termination only if not already terminating */
    // TODO: This should not be done if straw has already been terminated by the parent!
    TerminateStraw(Session_p, Straw_p);
  }

  return TRUE;
}


void
StrawAddMapping(StrawSession_t*             Session_p,
                const StrawSignalMapping_t* Mapping_p)
{
  unsigned int SignalHash;

  /* Get signal hash */
  SignalHash = CalcualteSignalHash(Mapping_p[0].InternalSignal, Mapping_p[0].ExternalSignal);

  /* Add straw to hash table */
  /* NOTE: Only the first signal is hashed. This means that a normal straw will get a speed boost,
   *       but a server straw will only get a boost for the first signal in the mapping. This
   *       solution is used to preserve lots of RAM and speed up the mapping removal process. The
   *       worst scenario for this solution is a server straw with thousands of other straws - in
   *       that case the mapping can become a true linears search. This should not matter much for
   *       server straws though, since their functions usually have other timing bottlenecks.
   */
  Session_p->CurrentStraw_p->SlotC.SignalMapping_p = Mapping_p;
  Session_p->CurrentStraw_p->SlotB.NextHashEntry_p = Session_p->MappingsHash_p[SignalHash];
  Session_p->CurrentStraw_p->Flags |= STRAW_WAITFORSIGNAL;
  Session_p->MappingsHash_p[SignalHash] = Session_p->CurrentStraw_p;
}
